require 'libxml'
require 'libxslt'
require 'mnogo'

mnogo = Mnogo.new('mysql://root:@localhost/mnogosearch/?dbmode=single','single')
# необходимо для правильной обработки полученного XML'я
mnogo.set_param(Mnogo::UDM_PARAM_CHARSET, "koi8-r")
if mnogo.version > 30204
mnogo.set_param(Mnogo::UDM_PARAM_BROWSER_CHARSET, "koi8-r")
end
mnogo.set_param(Mnogo::UDM_PARAM_HLBEG,'<hlight>')
mnogo.set_param(Mnogo::UDM_PARAM_HLEND,'</hlight>')
res = mnogo.find(ARGV[0]?ARGV[0]:"привод")
File.open('test.xml', 'w') do |f|
  f.write res.xml.gsub(/&/, '&amp;').gsub("\2",'<hlight>').gsub("\3",'</hlight>')
end

xslt = XML::XSLT.file('test.xsl')
xslt.doc = XML::Document.file('test.xml')
s = xslt.parse
s.apply

File.open('test.html', 'w') do |f|
  s.save(f)
end
