require 'mnogo'

mnogo = Mnogo.new('mysql://root@localhost/mnogosearch/','crc-multi')
# необходимо для правильной обработки полученного XML'я
mnogo.set_param(Mnogo::UDM_PARAM_CHARSET, "koi8-r")
if mnogo.version > 30204
mnogo.set_param(Mnogo::UDM_PARAM_BROWSER_CHARSET, "koi8-r")
end
#
mnogo.find(ARGV[0]?ARGV[0]:'Статьи').each { |x| print '<a href="',x.url,'">',x.title,"</a><br>\n" }
